///////////////////////////////////////////////////////////////////////////////
// File: ManagedHost.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "ManagedHost.h"
#include "Socket.h"

#include "ManagedHostInterface/IManagedHost.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/ToString.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/IProvideUserData.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/IOTools/IBufferBase.h"

#include "JetByteTools/COMTools/Utils.h"

#include "JetByteTools/CLRHostingTools/Exception.h"

#pragma hdrstop

#include <comdef.h>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IProvideUserData;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::Output;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ILockableObject;

using JetByteTools::Win32::CWin32Exception;

using JetByteTools::IO::IBufferBase;

using JetByteTools::Socket::IStreamSocket;

using JetByteTools::CLRHosting::CCLRHost;

using JetByteTools::COM::SafeAddRef;
using JetByteTools::COM::SafeRelease;

///////////////////////////////////////////////////////////////////////////////
// CManagedHost
///////////////////////////////////////////////////////////////////////////////

CManagedHost::CManagedHost(
   IProvideUserData &dataProvider,
   const _tstring &welcomeMessage,
   const _tstring &assembly,
   const _tstring &type)
   :  m_pClrHost(0),
      m_welcomeMessage(welcomeMessage),
      m_assembly(assembly),
      m_type(type),
      m_managerIndex(dataProvider.RequestUserDataSlot(_T("CManagedHost - Manager"))),
      m_appDomainIDIndex(dataProvider.RequestUserDataSlot(_T("CManagedHost - ID"))),
      m_managedSocketIndex(dataProvider.RequestUserDataSlot(_T("CManagedHost - Socket"))),
      m_managedDataIndex(dataProvider.RequestUserDataSlot(_T("CManagedHost - Data")))
{

}

CManagedHost::~CManagedHost()
{
   try
   {
      if (m_pClrHost)
      {
         m_pClrHost->DisableCallbacks(true);
      }

      for (DomainManagerMap::iterator it = m_domainManagers.begin(); it != m_domainManagers.end(); it++)
      {
         SafeRelease(it->second);
      }

      if (m_pClrHost)
      {
         m_pClrHost->Stop();
      }

      delete m_pClrHost;
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

void CManagedHost::Start()
{
   if (m_pClrHost)
   {
      throw CException(_T("CManagedHost::Start()"), _T("Already started"));
   }

   m_pClrHost = new CCLRHost(CCLRHost::Server, 0, this);

   m_pClrHost->SetAppDomainManagerType(m_assembly, m_type);

   m_pClrHost->Start();
}

void CManagedHost::Stop()
{
   if (!m_pClrHost)
   {
      throw CException(_T("CManagedHost::Start()"), _T("Not started"));
   }

   m_pClrHost->Stop();

   delete m_pClrHost;

   m_pClrHost = 0;
}

void CManagedHost::OnConnectionEstablished(
   const bool createPerConnectionApplicationDomain,
   IIndexedOpaqueUserData &userData,
   IStreamSocket &socket)
{
   IManagedHost *pDefaultManager = GetDefaultManagedHost();

   IManagedHost *pManager = pDefaultManager;

   int id = 0;

   if (createPerConnectionApplicationDomain)
   {
      _bstr_t name(L"Name");

      HRESULT hr = pDefaultManager->CreateAppDomain(name, &id);

      JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CManagedHost::OnConnectionEstablished()"), hr);

      pManager = GetManagedHost(id);
   }

   userData.SetUserPointer(m_managerIndex, pManager);
   userData.SetUserData(m_appDomainIDIndex, id);

   CSocket *pSocket = new CSocket(socket);

   userData.SetUserPointer(m_managedSocketIndex, pSocket);

   SAFEARRAY *pArray = 0;

   HRESULT hr = SafeArrayAllocDescriptorEx(VT_UI1, 1, &pArray);

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CManagedHost::OnConnectionEstablished() - SafeArrayAllocDescriptorEx"), hr);

   pArray->pvData = 0;
   pArray->rgsabound[0].lLbound = 0;
   pArray->rgsabound[0].cElements = 0;

   userData.SetUserPointer(m_managedDataIndex, pArray);

   pManager->OnConnectionEstablished(pSocket);
}

void CManagedHost::OnConnectionClientClose(
   IIndexedOpaqueUserData &userData)
{
   IManagedHost *pManager = GetManagedHost(userData);

   CSocket *pSocket = GetManagedSocket(userData);

   pManager->OnConnectionClientClose(pSocket);
}

void CManagedHost::ReadCompleted(
   IIndexedOpaqueUserData &userData,
   const IBufferBase &buffer)
{
   try
   {
      IManagedHost *pManager = GetManagedHost(userData);

      CSocket *pSocket = GetManagedSocket(userData);

      SAFEARRAY *pArray = GetManagedData(userData, buffer);

      HRESULT hr = pManager->OnReadCompleted(
         pArray,
         pSocket);

      if (FAILED(hr))
      {
         Output(_T("CManagedHost::ReadCompleted() - Managed call failed: ") + ErrorCodeToErrorMessage(hr));
      }
   }
   catch(const CException &e)
   {
      Output(_T("CManagedHost::ReadCompleted() - Exception: ") + e.GetDetails());
   }
   catch(...)
   {
      Output(_T("CManagedHost::ReadCompleted() - Unexpected exception"));
   }
}

void CManagedHost::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   const DWORD appDomainID = static_cast<DWORD>(userData.GetUserData(m_appDomainIDIndex));

   if (appDomainID != 0)
   {
      m_pClrHost->UnloadAppDomain(appDomainID);

      IManagedHost *pDomainManager = 0;

      {
         ILockableObject::Owner lock(m_lock);

         DomainManagerMap::iterator it = m_domainManagers.find(appDomainID);

         if (m_domainManagers.end() == it)
         {
            throw CException(_T("CManagedHost::OnSocketReleased()"), _T("Failed to locate AppDomain manager for id: ") + ToString(appDomainID));
         }

         pDomainManager = it->second;

         m_domainManagers.erase(it);
      }

      SafeRelease(pDomainManager);
   }

   DestroyManagedData(userData);
}

IManagedHost *CManagedHost::GetManagedHost(
   IIndexedOpaqueUserData &userData) const
{
   return reinterpret_cast<IManagedHost *>(userData.GetUserPointer(m_managerIndex));
}

CSocket *CManagedHost::GetManagedSocket(
   IIndexedOpaqueUserData &userData) const
{
   return reinterpret_cast<CSocket *>(userData.GetUserPointer(m_managedSocketIndex));
}

SAFEARRAY *CManagedHost::GetManagedData(
   IIndexedOpaqueUserData &userData,
   const IBufferBase &buffer) const
{
   SAFEARRAY *pArray = reinterpret_cast<SAFEARRAY *>(userData.GetUserPointer(m_managedDataIndex));

   pArray->pvData = const_cast<BYTE *>(buffer.GetMemory());
   pArray->rgsabound[0].cElements = buffer.GetUsed();

   return pArray;
}

void CManagedHost::DestroyManagedData(
   IIndexedOpaqueUserData &userData) const
{
   reinterpret_cast<CSocket *>(userData.GetUserPointer(m_managedSocketIndex))->Disconnect();

   userData.SetUserPointer(m_managedSocketIndex, 0);

   SAFEARRAY *pArray = reinterpret_cast<SAFEARRAY *>(userData.GetUserPointer(m_managedDataIndex));

   userData.SetUserPointer(m_managedDataIndex, 0);

   SafeArrayDestroyDescriptor(pArray);
}

IManagedHost *CManagedHost::GetManagedHost(
   const long appDomainID) const
{
   IManagedHost *pHost = 0;

   ILockableObject::Owner lock(m_lock);

   DomainManagerMap::const_iterator it = m_domainManagers.find(appDomainID);

   if (m_domainManagers.end() != it)
   {
      pHost = SafeAddRef(it->second);
   }

#pragma JETBYTE_TODO("return smart pointers")

   if (!pHost)
   {
      throw CException(_T("CManagedHost::GetManagedHost()"), _T("AppDomain Manager not available for id: ") + ToString(appDomainID));
   }

   return pHost;
}

IManagedHost *CManagedHost::GetDefaultManagedHost() const
{
   return GetManagedHost(1);
}

HRESULT CManagedHost::GetHostManager(
   REFIID /*riid*/,
   void **ppObject)
{
   if (!ppObject)
   {
      return E_POINTER;
   }

   *ppObject = 0;

   HRESULT hr = E_NOINTERFACE;

   return hr;
}

HRESULT CManagedHost::SetAppDomainManager(
  DWORD dwAppDomainID,
  IUnknown *pUnkAppDomainManager)
{
   ILockableObject::Owner lock(m_lock);

   DomainManagerMap::iterator it = m_domainManagers.find(dwAppDomainID);

   if (m_domainManagers.end() != it)
   {
      it->second = SafeRelease(it->second);
   }

   IManagedHost *pHost = 0;

   HRESULT hr = pUnkAppDomainManager->QueryInterface(IID_IManagedHost, (void**)&pHost);

   m_domainManagers[dwAppDomainID] = pHost;

   return hr;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ManagedHost.cpp
///////////////////////////////////////////////////////////////////////////////
