///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "SocketServer.h"
#include "ManagedHost.h"
#include "IOPool.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/CommandLine.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/ServerCollection.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"
#include "JetByteTools/SocketTools/FullAddress.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Socket::CServerCollection;
using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   ::CoInitializeEx(0, COINIT_MULTITHREADED);

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("EchoServerCLR.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CCommandLine commandLine(argc, argv, _T("EchoServerCLR"), CCommandLine::TCPServer);

         if (commandLine.Parse())
         {
            CIOPool pool(commandLine.NumberOfIOThreads());

            pool.Start();

            CSequencedStreamSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            const ListenBacklog listenBacklog = commandLine.ListenBacklog();

            CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

            const _tstring assembly = _T("ManagedHost, Version=1.0.0.0, PublicKeyToken=24e328fd91cb8ee6");
            const _tstring type = _T("EchoServerCLR.ManagedHost");

            const _tstring welcomeMessage = commandLine.NoWelcomeMessage() ? _T("") : _T("Welcome to echo server\r\n");

            CManagedHost managedHost(
               socketAllocator,
               welcomeMessage,
               assembly,
               type);

            managedHost.Start();

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address1(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            CServerCollection servers;

            servers.AddServer(new CSocketServer(
               managedHost,
               false,
               address1,
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            const CFullAddress address2(
               commandLine.Server(),
               commandLine.Port() + 1,
               addressPreference);

            servers.AddServer(new CSocketServer(
               managedHost,
               true,
               address2,
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            servers.Start();

            servers.StartAcceptingConnections();

            shutdownHandler.SetServerControl(servers, commandLine.EventNamePrefix());

            shutdownHandler.WaitForShutdownRequest();

            servers.WaitForShutdownToComplete();

            pool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();

            managedHost.Stop();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
