///////////////////////////////////////////////////////////////////////////////
// File: AssemblyManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "AssemblyManager.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/ToString.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/Win32Tools/DirectorySearch.h"
#include "JetByteTools/Win32Tools/Win32Exception.h"
#include "JetByteTools/Win32Tools/SmartHandle.h"

#include "JetByteTools/CLRHostingTools/Exception.h"

#pragma hdrstop

#include <algorithm>    // for min

#include <limits>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::StringVector;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::GetCurrentDirectory;
using JetByteTools::Core::ILockableObject;

using JetByteTools::Win32::CDirectorySearch;
using JetByteTools::Win32::CWin32Exception;
using JetByteTools::Win32::CSmartHandle;

using JetByteTools::CLRHosting::CException;

using std::wstring;
using std::min;
using std::numeric_limits;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static IStream *CopyIStream(
   IStream *pInputStream);

static IStream *LoadFileAsIStreamOnHGlobal(
   const _tstring &fileName,
   const ULONGLONG fileSize);

///////////////////////////////////////////////////////////////////////////////
// CManagedHost
///////////////////////////////////////////////////////////////////////////////

CAssemblyManager::CAssemblyManager(
   const _tstring &assemblyStorePath,
   const StringVector &nonHostAssemblies)
   :  CHostAssemblyManager(nonHostAssemblies),
      m_assemblyStorePath(assemblyStorePath),
      m_nextAssemblyId(1)
{
   try
   {
      OutputEx(_T("Current Dir = ") + GetCurrentDirectory());
      OutputEx(_T("Assembly Store Path = ") + m_assemblyStorePath);

      CDirectorySearch search(m_assemblyStorePath + _T("\\*.dll"));

      ICLRAssemblyIdentityManager *pIdentityManager = GetAssemblyIdentityManager();

      while(search.HasFile())
      {
         OutputEx(search.GetFileName());

         const wstring fileName = CStringConverter::TtoW(search.GetFileName());

         wstring buffer;

         DWORD dataLength = 0;

         HRESULT hr = S_OK;

         do
         {
            hr = pIdentityManager->GetBindingIdentityFromFile(
               fileName.c_str(),
               CLR_ASSEMBLY_IDENTITY_FLAGS_DEFAULT,
               const_cast<wchar_t *>(buffer.c_str()),
               &dataLength);

            if (FAILED(hr))
            {
               if (hr != HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER))
               {
                  throw CException(_T("CAssemblyManager::CAssemblyManager() - GetBindingIdentityFromFile"), hr);
               }

               buffer.resize(dataLength - 1);
            }
         }
         while (FAILED(hr));

         OutputEx(buffer);

         {
            const _tstring fileName = search.GetFileName();

            search.NextFile();

            IStream *pIStream = LoadFileAsIStreamOnHGlobal(fileName, search.GetFileSize());

            ILockableObject::Owner lock(m_lock);

            m_assemblies[buffer] = AssemblyDetails(pIStream, m_nextAssemblyId++);
         }
      }

      pIdentityManager->Release();
   }
   catch(CException &)
   {

   }
}

CAssemblyManager::~CAssemblyManager()
{
   try
   {
      for (Assemblies::const_iterator it = m_assemblies.begin(); it != m_assemblies.end(); ++it)
      {
         it->second.first->Release();
      }
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

HRESULT CAssemblyManager::ProvideAssembly(
   AssemblyBindInfo *pBindInfo,
   UINT64 *pAssemblyId,
   UINT64 *pContext,
   IStream **ppStmAssemblyImage,
   IStream **ppStmPDB)
{
   if (!ppStmAssemblyImage || !ppStmPDB)
   {
      return E_POINTER;
   }

   *pAssemblyId = 0;
   *pContext = 0;
   *ppStmAssemblyImage = 0;
   *ppStmPDB = 0;

   OutputEx(_T("CAssemblyManager::ProvideAssembly()"));
   OutputEx(_T("AppDomain: ") + ToString(pBindInfo->dwAppDomainId));
   OutputEx(_T("Reference Identity: \"") + CStringConverter::WtoT(pBindInfo->lpReferencedIdentity) + _T("\""));
   OutputEx(_T("Post Policy Identity: \"") + CStringConverter::WtoT(pBindInfo->lpPostPolicyIdentity) + _T("\""));
   OutputEx(_T("PolicyLevel: ") + ToString(pBindInfo->ePolicyLevel));

   ILockableObject::Owner lock(m_lock);

   Assemblies::const_iterator it = m_assemblies.find(pBindInfo->lpPostPolicyIdentity);

   if (it != m_assemblies.end())
   {
      IStream *pIStream = CopyIStream(it->second.first);

      *ppStmAssemblyImage = pIStream;

      *pAssemblyId = it->second.second;

      *pContext = 42;

      OutputEx(_T("Found and loaded"));

      return S_OK;
   }

   return HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
}

HRESULT CAssemblyManager::ProvideModule(
   ModuleBindInfo *pBindInfo,
   DWORD *pdwModuleId,
   IStream **ppStmModuleImage,
   IStream **ppStmPDB)
{
   if (!ppStmModuleImage || !ppStmPDB)
   {
      return E_POINTER;
   }

   *pdwModuleId = 0;

   Output(_T("CAssemblyManager::ProvideModule()"));
   Output(_T("AppDomain: ") + ToString(pBindInfo->dwAppDomainId));
   Output(_T("Assembly Identity: \"") + CStringConverter::WtoT(pBindInfo->lpAssemblyIdentity) + _T("\""));
   Output(_T("Module Name: \"") + CStringConverter::WtoT(pBindInfo->lpModuleName) + _T("\""));

   return HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static IStream *CopyIStream(
   IStream *pInputStream)
{
   IStream *pOutputStream = 0;

   HRESULT hr = pInputStream->Clone(&pOutputStream);

   CException::ThrowOnFailure(_T("CopyIStream()"), hr);

   return pOutputStream;
}

static IStream *LoadFileAsIStreamOnHGlobal(
   const _tstring &fileName,
   const ULONGLONG fileSize)
{
   HGLOBAL hGlobal = ::GlobalAlloc(GHND, static_cast<size_t>(fileSize));

   CSmartHandle hFile(::CreateFile(
      fileName.c_str(),
      GENERIC_READ,
      FILE_SHARE_READ,
      0,
      OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL,
      0));

   if (hFile == INVALID_HANDLE_VALUE)
   {
      const DWORD lastError = ::GetLastError();

      throw CWin32Exception(_T("LoadFileAsIStreamOnHGlobal() - CreateFile() \"") + fileName + _T("\""), lastError);
   }

   BYTE *pData = reinterpret_cast<BYTE *>(::GlobalLock(hGlobal));

   bool done = false;

   DWORD totalBytesRead = 0;

   do
   {
      DWORD bytesRead = 0;

      DWORD bytesToRead = static_cast<DWORD>(min(static_cast<ULONGLONG>(fileSize - totalBytesRead), static_cast<ULONGLONG>(numeric_limits<DWORD>::max())));

      if (::ReadFile(hFile, pData + totalBytesRead, bytesToRead, &bytesRead, 0) && bytesRead > 0)
      {
         totalBytesRead += bytesRead;
      }
      else
      {
         done = true;
      }
   }
   while (!done);

   if (!::GlobalUnlock(hGlobal))
   {
      const DWORD lastError = ::GetLastError();

      if (lastError != NO_ERROR)
      {
         throw CWin32Exception(_T("LoadFileAsIStreamOnHGlobal() - GlobalUnlock() \"") + fileName + _T("\""), lastError);
      }
   }

   IStream *pStream = 0;

   HRESULT hr = ::CreateStreamOnHGlobal(hGlobal, TRUE, &pStream);

   CException::ThrowOnFailure(_T("LoadFileAsIStreamOnHGlobal() - CreateStreamOnHGlobal() \"") + fileName + _T("\""), hr);

   return pStream;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: CAssemblyManager.cpp
///////////////////////////////////////////////////////////////////////////////
