///////////////////////////////////////////////////////////////////////////////
// File: Application.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2010 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "Application.h"
#include "ApplicationCollection.h"

#include "ManagedHostInterface/IManagedHost.h"

#include "JetByteTools/CoreTools/Exception.h"

#pragma hdrstop

#include "JetByteTools/SocketTools/IStreamSocket.h"

#include "JetByteTools/COMTools/Utils.h"

#include "JetByteTools/CLRHostingTools/Exception.h"

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::ILockableObject;

using JetByteTools::Socket::IStreamSocket;

using JetByteTools::COM::SafeAddRef;
using JetByteTools::COM::SafeRelease;

///////////////////////////////////////////////////////////////////////////////
// CApplication
///////////////////////////////////////////////////////////////////////////////

CApplication::CApplication(
   const _tstring &name,
   const long appDomainID,
   IManagedApplication *pApp,
   CApplicationCollection &collection)
   :  m_ref(1),
      m_name(name),
      m_appDomainID(appDomainID),
      m_pApp(SafeAddRef(pApp)),
      m_collection(collection)
{
   if (!m_pApp)
   {
      throw CException(_T("CApplication::CApplication()"), _T("Unexpected: Application pointer is null"));
   }
}

CApplication::~CApplication()
{
   try
   {
      SafeRelease(m_pApp);

      m_collection.OnApplicationDestroyed(m_appDomainID);
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

const _tstring &CApplication::GetName() const
{
   return m_name;
}

void CApplication::Connect(
   IStreamSocket &socket)
{
   ILockableObject::Owner lock(m_lock);

   Connections::const_iterator it = m_connections.find(&socket);

   if (it != m_connections.end())
   {
      throw CException(_T("CApplication::Connect()"), _T("This socket is already connected to this application"));
   }

   m_connections.insert(&socket);

   AddRef();
}

void CApplication::Disconnect(
   IStreamSocket &socket)
{
   {
      ILockableObject::Owner lock(m_lock);

      Connections::iterator it = m_connections.find(&socket);

      if (it == m_connections.end())
      {
         throw CException(_T("CApplication::Disconnect()"), _T("This socket is not connected to this application"));
      }

      m_connections.erase(it);
   }

   // Lock must have been released before we call release or we might try and unlock
   // after we've been deleted.

   Release();
}

void CApplication::AbortAllConnections()
{
   ILockableObject::Owner lock(m_lock);

   for (Connections::const_iterator it = m_connections.begin(), end = m_connections.end();
      it != end;
      ++it)
   {
      IStreamSocket *pSocket = *it;

      pSocket->AbortConnection();
   }
}

void CApplication::AddRef()
{
   ::InterlockedIncrement(&m_ref);
}

void CApplication::Release()
{
   if (0 == ::InterlockedDecrement(&m_ref))
   {
      delete this;
   }
}

long CApplication::GetAppDomainID() const
{
   return m_appDomainID;
}

void CApplication::OnConnectionEstablished(
   ISocket &socket)
{
   HRESULT hr = m_pApp->OnConnectionEstablished(&socket);

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CApplication::OnConnectionEstablished()"), hr);
}

void CApplication::OnConnectionClientClose(
   ISocket &socket)
{
   HRESULT hr = m_pApp->OnConnectionClientClose(&socket);

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CApplication::OnConnectionClientClose()"), hr);
}

void CApplication::OnReadCompleted(
   ISocket &socket,
   SAFEARRAY &data)
{
   HRESULT hr = m_pApp->OnReadCompleted(&data, &socket);

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CApplication::OnReadCompleted()"), hr);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: Application.cpp
///////////////////////////////////////////////////////////////////////////////
