///////////////////////////////////////////////////////////////////////////////
// File: ApplicationCollection.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2010 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "ApplicationCollection.h"
#include "Application.h"

#include "ManagedHostInterface/IManagedHost.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/ToString.h"
#include "JetByteTools/CoreTools/ReferenceCountedSmartPointer.h"

#include "JetByteTools/COMTools/Utils.h"

#pragma hdrstop

#include <comdef.h>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::ILockableObject;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CThreadedCallbackTimerQueue;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::OutputEx;

using JetByteTools::Win32::CDirectoryChangeMonitor;

using JetByteTools::COM::SafeAddRef;
using JetByteTools::COM::SafeRelease;

///////////////////////////////////////////////////////////////////////////////
// Typedefs
///////////////////////////////////////////////////////////////////////////////

typedef JetByteTools::Core::TReferenceCountedSmartPointer<IManagedApplication> CSmartManagedApplication;

typedef JetByteTools::Core::TReferenceCountedSmartPointer<CApplicationCollection::ApplicationDetails> CSmartApplicationDetails;

///////////////////////////////////////////////////////////////////////////////
// CApplicationCollection
///////////////////////////////////////////////////////////////////////////////

CApplicationCollection::CApplicationCollection(
   const _tstring &baseDirectory,
   const _tstring &assembly,
   const _tstring &type,
   const bool enableManagedDebuggingAssistants)
   :  m_baseDirectory(baseDirectory),
      m_host(assembly, type, *this, enableManagedDebuggingAssistants)
{

}

CApplicationCollection::~CApplicationCollection()
{
   try
   {
      RemoveAllApplications();
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

void CApplicationCollection::AddApplication(
   const _tstring &application,
   const _tstring &applicationDirectory,
   const _tstring &sharedDirectory,
   const _tstring &configFile,
   const AssemblyReloadFlags assemblyReloadFlags,
   const _tstring &assembly,
   const _tstring &type)
{
   ILockableObject::Owner lock(m_lock);

   ApplicationDetailsMap::const_iterator it = m_applicationDetails.find(application);

   if (it != m_applicationDetails.end())
   {
      throw CException(_T("CApplicationCollection::AddApplication("), _T("Duplicate application name: \"") + application + _T("\""));
   }

   ApplicationDetails *pAppDetails = new ApplicationDetails(
      m_directoryChangeMonitor,
      m_timerQueue,
      *this,
      m_baseDirectory,
      application,
      applicationDirectory,
      sharedDirectory,
      configFile,
      assemblyReloadFlags,
      assembly,
      type);

   m_applicationDetails.insert(ApplicationDetailsMap::value_type(application, pAppDetails));
}

bool CApplicationCollection::StartApplication(
   const _tstring &application)
{
   bool ok = false;

   ILockableObject::Owner lock(m_lock);

   ApplicationDetails &appDetails = GetApplicationDetails(application);

   ApplicationByNameMap::const_iterator it = m_applicationsByName.find(application);

   if (it == m_applicationsByName.end())
   {
      const long appDomainID = m_host.CreateApplicationInNewAppDomain(
         application,
         appDetails.applicationDirectory,
         appDetails.sharedDirectory,
         appDetails.configFile,
         appDetails.assemblyReloadFlags != DisableShadowCopy);

      try
      {
         appDetails.BeginMonitoring();

         CSmartManagedApplication spManagedApp(m_host.StartApplication(appDomainID, appDetails.assembly, appDetails.type, application));

         CSmartApplication spApplication(new CApplication(application, appDomainID, spManagedApp.Get(), *this));

         m_applicationsByName.insert(ApplicationByNameMap::value_type(application, spApplication.Get()));

         m_applicationsById.insert(ApplicationByIdMap::value_type(appDomainID, spApplication.Detach()));

         ok = true;
      }
      catch(...)
      {
         m_host.DestroyAppDomain(appDomainID);

         appDetails.EndMonitoring();

         throw;
      }
   }

   return ok;
}

bool CApplicationCollection::RestartApplication(
   const _tstring &application,
   const bool kill)
{
   bool ok = false;

   ILockableObject::Owner lock(m_lock);

   ApplicationDetails &appDetails = GetApplicationDetails(application);

   const long appDomainID = m_host.CreateApplicationInNewAppDomain(
      application,
      appDetails.applicationDirectory,
      appDetails.sharedDirectory,
      appDetails.configFile,
      appDetails.assemblyReloadFlags != DisableShadowCopy);

   try
   {
      appDetails.BeginMonitoring();

      CSmartManagedApplication spManagedApp(m_host.StartApplication(appDomainID, appDetails.assembly, appDetails.type, application));

      CSmartApplication spApplication(new CApplication(application, appDomainID, spManagedApp.Get(), *this));

      ApplicationByNameMap::iterator it = m_applicationsByName.find(application);

      if (it != m_applicationsByName.end())
      {
         CApplication *pApp = it->second;

         StopApplication(*pApp, kill);

         it->second = spApplication.Get();

         m_applicationsById.insert(ApplicationByIdMap::value_type(appDomainID, spApplication.Detach()));

         ok = true;
      }
      else
      {
         m_applicationsByName.insert(ApplicationByNameMap::value_type(application, spApplication.Get()));

         m_applicationsById.insert(ApplicationByIdMap::value_type(appDomainID, spApplication.Detach()));
      }
   }
   catch(...)
   {
      m_host.DestroyAppDomain(appDomainID);

      appDetails.EndMonitoring();

      throw;
   }

   return ok;
}

bool CApplicationCollection::ApplicationIsRunning(
   const _tstring &application) const
{
   bool running = false;

   ILockableObject::Owner lock(m_lock);

   ValidateApplication(application);

   CSmartApplication spApplication(GetApplication(application));

   if (spApplication.Get())
   {
      running = true;
   }

   return running;
}

void CApplicationCollection::ValidateApplication(
   const _tstring &application) const
{
   GetApplicationDetails(application);
}

CApplicationCollection::ApplicationDetails &CApplicationCollection::GetApplicationDetails(
   const _tstring &application)
{
   ApplicationDetailsMap::const_iterator it = m_applicationDetails.find(application);

   if (it == m_applicationDetails.end())
   {
      throw CException(_T("CApplicationCollection::GetApplicationDetails("), _T("Application: \"") + application + _T("\" not found"));
   }

   ApplicationDetails *pAppDetails = it->second;

   return *pAppDetails;
}

const CApplicationCollection::ApplicationDetails &CApplicationCollection::GetApplicationDetails(
   const _tstring &application) const
{
   ApplicationDetailsMap::const_iterator it = m_applicationDetails.find(application);

   if (it == m_applicationDetails.end())
   {
      throw CException(_T("CApplicationCollection::GetApplicationDetails("), _T("Application: \"") + application + _T("\" not found"));
   }

   ApplicationDetails *pAppDetails = it->second;

   return *pAppDetails;
}

CSmartApplication CApplicationCollection::GetApplication(
   const _tstring &application) const
{
   ILockableObject::Owner lock(m_lock);

   ApplicationByNameMap::const_iterator it = m_applicationsByName.find(application);

   if (it != m_applicationsByName.end())
   {
      return CSmartApplication(it->second, CSmartApplication::CreateNewReference);
   }

   return CSmartApplication(0);
}

bool CApplicationCollection::StopApplication(
   const _tstring &application,
   const bool kill)
{
   bool ok = false;

   ILockableObject::Owner lock(m_lock);

   ApplicationByNameMap::iterator it = m_applicationsByName.find(application);

   if (it != m_applicationsByName.end())
   {
      ApplicationDetails &appDetails = GetApplicationDetails(application);

      appDetails.EndMonitoring();

      CApplication *pApp = it->second;

      StopApplication(*pApp, kill);

      m_applicationsByName.erase(it);

      // Application cannot be connected to by new connections
      // when last active reference is release it will stop and
      // the app domain will be destroyed.

      ok = true;
   }

   return ok;
}

void CApplicationCollection::RemoveApplication(
   const _tstring &application,
   const bool kill)
{
   ILockableObject::Owner lock(m_lock);

   ApplicationDetailsMap::iterator it = m_applicationDetails.find(application);

   if (it == m_applicationDetails.end())
   {
      throw CException(_T("CApplicationCollection::RemoveApplication("), _T("Application: \"") + application + _T("\" not found"));
   }

   delete it->second;

   m_applicationDetails.erase(it);

   StopApplication(application, kill);
}

void CApplicationCollection::StopAllApplications(
   const bool kill)
{
   ILockableObject::Owner lock(m_lock);

   for (ApplicationByNameMap::const_iterator it = m_applicationsByName.begin(), end = m_applicationsByName.end();
      it != end;
      ++it)
   {
      CApplication *pApp = it->second;

      StopApplication(*pApp, kill);

      // Application cannot be connected to by new connections
      // when last active reference is release it will stop and
      // the app domain will be destroyed.
   }

   m_applicationsByName.clear();

   m_applicationsById.clear();
}

void CApplicationCollection::RemoveAllApplications(
   const bool kill)
{
   ILockableObject::Owner lock(m_lock);

   for (ApplicationDetailsMap::const_iterator it = m_applicationDetails.begin(), end = m_applicationDetails.end();
      it != end;
      ++it)
   {
      ApplicationDetails *pAppDetails = it->second;

      pAppDetails->Release();
   }

   m_applicationDetails.clear();

   StopAllApplications(kill);
}

CApplicationCollection::Iterator CApplicationCollection::Begin() const
{
   return m_applicationDetails.begin();
}

CApplicationCollection::Iterator CApplicationCollection::End() const
{
   return m_applicationDetails.end();
}

void CApplicationCollection::OnApplicationDestroyed(
   const long appDomainID)
{
   try
   {
      m_host.StopApplication(appDomainID);

      m_host.DestroyAppDomain(appDomainID);
   }
   catch(...)
   {
      // if the app domain was forcibly unloaded due to something
      // like a stack overflow then the host cant stop or destroy it...
      // in fact it's already been removed from the host
      // there should be a better way to do this
   }
}

void CApplicationCollection::StopApplication(
   CApplication &application,
   const bool kill)
{
   m_applicationsById.erase(application.GetAppDomainID());

   if (kill)
   {
      application.AbortAllConnections();
   }
   else
   {
      m_host.RequestApplicationStop(application.GetAppDomainID());
   }

   application.Release();
}

void CApplicationCollection::OnAppDomainUnload(
   const long appDomainID)
{
   ILockableObject::Owner lock(m_lock);

   ApplicationByIdMap::iterator it = m_applicationsById.find(appDomainID);

   if (it != m_applicationsById.end())
   {
      CApplication *pApp = it->second;

      m_applicationsByName.erase(pApp->GetName());

      m_applicationsById.erase(it);

      pApp->Release();
   }
}

void CApplicationCollection::OnTimer(
   UserData userData)
{
   try
   {
      ILockableObject::Owner lock(m_lock);

      CSmartApplicationDetails spAppDetails(reinterpret_cast<ApplicationDetails *>(userData));

      ApplicationDetailsMap::const_iterator it = m_applicationDetails.find(spAppDetails->application);

      if (it != m_applicationDetails.end())
      {
         if (spAppDetails->assemblyReloadFlags == EnableAutoRestart ||
             spAppDetails->assemblyReloadFlags == EnableAutoForceRestart)
         {
            const bool kill = (spAppDetails->assemblyReloadFlags == EnableAutoForceRestart);

            OutputEx(_T("Restarting application: \"") + spAppDetails->application + _T("\" due to file changes"));

            RestartApplication(spAppDetails->application, kill);
         }
         else
         {
            OutputEx(_T("Not restarting application: \"") + spAppDetails->application + _T("\" due to file changes, application cannot be restarted"));
         }
      }
      else
      {
         OutputEx(_T("Not restarting application: \"") + spAppDetails->application + _T("\" due to file changes, application not found"));
      }
   }
   catch(CException &e)
   {
      OutputEx(_T("Failed to restart application: ") + e.GetDetails());
   }
   catch(...)
   {
      OutputEx(_T("Failed to restart application: unexpected exception"));
   }
}

///////////////////////////////////////////////////////////////////////////////
// CApplicationCollection::ApplicationDetails
///////////////////////////////////////////////////////////////////////////////

CApplicationCollection::ApplicationDetails::ApplicationDetails(
   CDirectoryChangeMonitor &monitor,
   CThreadedCallbackTimerQueue &timerQueue,
   IQueueTimers::Timer &timer,
   const _tstring &baseDirectory,
   const _tstring &application_,
   const _tstring &applicationDirectory_,
   const _tstring &sharedDirectory_,
   const _tstring &configFile_,
   const AssemblyReloadFlags assemblyReloadFlags_,
   const _tstring &assembly_,
   const _tstring &type_)
   :  application(application_),
      applicationDirectory(applicationDirectory_),
      sharedDirectory(sharedDirectory_),
      configFile(configFile_),
      assemblyReloadFlags(assemblyReloadFlags_),
      assembly(assembly_),
      type(type_),
      m_monitor(monitor),
      m_timerQueue(timerQueue),
      m_timerHandle(timerQueue.CreateTimer()),
      m_timer(timer),
      m_baseDirectory(baseDirectory),
      m_ref(1)
{

}

CApplicationCollection::ApplicationDetails::~ApplicationDetails()
{
   try
   {
      OutputEx(_T("~ApplicationDetails()"));

      EndMonitoring();

      m_timerQueue.DestroyTimer(m_timerHandle);
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

void CApplicationCollection::ApplicationDetails::AddRef()
{
   ::InterlockedIncrement(&m_ref);
}

void CApplicationCollection::ApplicationDetails::Release()
{
   if (0 == ::InterlockedDecrement(&m_ref))
   {
      delete this;
   }
}

void CApplicationCollection::ApplicationDetails::BeginMonitoring()
{
   if (assemblyReloadFlags == CApplicationCollection::EnableAutoRestart ||
       assemblyReloadFlags == CApplicationCollection::EnableAutoForceRestart)
   {
      const _tstring directoryName = m_baseDirectory + _T("\\") + applicationDirectory;

      m_monitor.AddWatch(directoryName, FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME | FILE_NOTIFY_CHANGE_LAST_WRITE, *this, 0);

      // we should also walk the list of shared directories and add a watch to each of them as well
   }
}

void CApplicationCollection::ApplicationDetails::EndMonitoring()
{
   for (DirectoryChangeHandles::const_iterator it = m_handles.begin(), end = m_handles.end();
      it != end;
      ++it)
   {
       m_monitor.RemoveWatch(*it);
   }

   m_handles.clear();

   if (m_timerQueue.CancelTimer(m_timerHandle))
   {
      Release();
   }
}


void CApplicationCollection::ApplicationDetails::OnChange(
   const _tstring &directoryName,
   const DWORD action,
   const _tstring &fileName,
   const DWORD userData)
{
   // Lots of things we could do here such as check for and add a watch to new directories. Remove a watch from deleted directories, etc.
   (void)userData;
   (void)fileName;
   (void)action;

   JetByteTools::Core::OutputEx(_T("ApplicationDetails::OnChange() - ") + directoryName + _T(" - ") + fileName + _T(" - ") + JetByteTools::Core::ToString(action));

   AddRef();

   if (m_timerQueue.SetTimer(m_timerHandle, m_timer, 1000, reinterpret_cast<IQueueTimers::UserData>(this)))
   {
      Release();
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ApplicationCollection.cpp
///////////////////////////////////////////////////////////////////////////////
