///////////////////////////////////////////////////////////////////////////////
// File: ConfigurationSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2010 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "ConfigurationSocketServer.h"
#include "ApplicationCollection.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/Win32Tools/SecurityAttributes.h"
#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/ManualResetEvent.h"
#include "JetByteTools/Win32Tools/GlobalName.h"

#include "JetByteTools/IOTools/SmartBuffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::Output;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::GetStringLength;
using JetByteTools::Core::StripLeadingA;
using JetByteTools::Core::StripTrailingA;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::ToUpperA;
using JetByteTools::Core::IsAllDigitsA;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::OutputEx;

using JetByteTools::Win32::CSecurityAttributes;
using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CManualResetEvent;
using JetByteTools::Win32::CGlobalName;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::CStreamSocketServer;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static string ProcessCommandString(
   const string &input);

///////////////////////////////////////////////////////////////////////////////
// CConfigurationSocketServer
///////////////////////////////////////////////////////////////////////////////

CConfigurationSocketServer::CConfigurationSocketServer(
   CApplicationCollection &applications,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServer(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_applications(applications),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CSocketServer - m_userDataIndex")))
{

}

CConfigurationSocketServer::~CConfigurationSocketServer()
{
   try
   {
      WaitForShutdownToComplete();
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

void CConfigurationSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   if (socket.TryRead())
   {
      static const string signOnMessage =
         "Welcome to the configuration server\r\n"
         "Type help and press return to list available command\r\n"
         "->";

      socket.TryWrite(signOnMessage.c_str(), static_cast<DWORD>(signOnMessage.length()));
   }
}

void CConfigurationSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      ProcessDataStream(socket, buffer);

      socket.TryRead(buffer);
   }
   catch(const CException &e)
   {
      Output(_T("OnReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      Output(_T("OnReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CConfigurationSocketServer::ProcessDataStream(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   bool done;

   do
   {
      done = true;

      const IBuffer::BufferSize used = buffer->GetUsed();

      if (used >= GetMinimumMessageSize())
      {
         const IBuffer::BufferSize startOffset = static_cast<IBuffer::BufferSize>(socket.GetUserData(m_userDataIndex));

         const IBuffer::BufferSize messageSize = GetMessageSize(buffer.GetRef(), startOffset);

         socket.SetUserData(m_userDataIndex, used);

         if (messageSize == 0)
         {
            // havent got a complete message yet.

            // we null terminate our messages in the buffer, so we need to reserve
            // a byte of the buffer for this purpose...

            if (used == (buffer->GetSize() - 1))
            {
               Output(_T("Too much data!"));

               const std::string response("Error: invalid command, too much data! Go away!\r\n");

               // Write this message and then shutdown the sending side of the socket.
               socket.Write(response.c_str(), GetStringLength<DWORD>(response));
               socket.Shutdown();

               // throw the rubbish away
               buffer->Empty();

               done = true;
            }
         }
         else if (used == messageSize)
         {
            buffer->AddData(0);   // null terminate the command string;

            ProcessCommand(socket, buffer);

            socket.SetUserData(m_userDataIndex, 0);

            if (buffer.Get())
            {
               buffer->Empty();
            }

            done = true;
         }
         else if (used > messageSize)
         {
            // allocate a new buffer, copy the data that we want into it and then read into the
            // original buffer again...

            CSmartBuffer message(buffer->SplitBufferFromFront(messageSize));

            message->AddData(0);   // null terminate the command string;

            ProcessCommand(socket, message);

            socket.SetUserData(m_userDataIndex, 0);

            // loop again, we may have another complete message in there...

            done = false;
         }
      }
   }
   while (!done);
}

IBuffer::BufferSize CConfigurationSocketServer::GetMinimumMessageSize() const
{
   // The smallest possible command we accept is TOP (plus the crlf terminator,
   // once we have this many bytes we can start with try and work out
   // what we have...
   return 2;
}

IBuffer::BufferSize CConfigurationSocketServer::GetMessageSize(
   const IBuffer &buffer,
   const IBuffer::BufferSize startOffset) const
{
   const BYTE *pData = buffer.GetMemory();

   const IBuffer::BufferSize used = buffer.GetUsed();

   for (IBuffer::BufferSize i = startOffset; i < used; ++i)
   {
      if (pData[i] == '\r')
      {
         if (i + 1 < used && pData[i + 1] == '\n')
         {
            // the end of the message is i+1
            // we actually want a count of characters, not a zero based
            // index, so we have to add 1...
            return i + 1 + 1;
         }
      }
   }

   return 0;
}

void CConfigurationSocketServer::ProcessCommand(
   IStreamSocket &socket,
   const CSmartBuffer &buffer)
{
   const string originalCommand = reinterpret_cast<const char*>(buffer->GetMemory());

   const string echoCommand = ProcessCommandString(originalCommand);

   string command = ToUpperA(echoCommand);

   const string::size_type length = command.length();

   bool ok = false;

   if (length == 0)
   {
      // just a \r\n
      ok = true;
   }
   else if (length >= 3)
   {
      if (length == 4)
      {
         if (command == "QUIT")
         {
            socket.Shutdown();

            return;
         }
         else if (command == "HELP")
         {
            ShowHelp(socket);

            ok = true;
         }
         else if (command == "LIST")
         {
            ListApplications(socket);

            ok = true;
         }
      }
      else
      {
         const string cmd = command.substr(0, 5);

         if (cmd == "STOP ")
         {
            ok = StopApplication(command.substr(5), socket, false);
         }
         else if (cmd == "KILL ")
         {
            ok = StopApplication(command.substr(5), socket, true);
         }
         else
         {
            const string cmd = command.substr(0, 6);

            if (cmd == "START ")
            {
               ok = StartApplication(command.substr(6), socket);
            }
            else
            {
               const string cmd = command.substr(0, 8);

               if (cmd == "RESTART ")
               {
                  ok = RestartApplication(command.substr(8), socket, false);
               }
               else if (cmd == "KISTART ")
               {
                  ok = RestartApplication(command.substr(8), socket, true);
               }
               else if (cmd == "SHUTDOWN")
               {
                  CSecurityDescriptorAllowAll sd;

                  CSecurityAttributes sa(sd);

                  CManualResetEvent shutdownEvent(CGlobalName(_T("JetByteToolsServerShutdown")), &sa);

                  shutdownEvent.Set();

                  ok = true;
               }
            }
         }
      }
   }

   if (!ok)
   {
      const string response =
         "Error: Invalid command: \"" + echoCommand + "\"\r\n"
         "For help type HELP and press return\r\n";

      socket.Write(response.c_str(), GetStringLength<DWORD>(response));

      OutputEx(DumpData(reinterpret_cast<const BYTE *>(echoCommand.c_str()), echoCommand.length()));
   }

   static const string prompt("->");


   socket.Write(prompt.c_str(), GetStringLength<DWORD>(prompt));
}

void CConfigurationSocketServer::ShowHelp(
   IStreamSocket &socket) const
{
   const static string help =
      "The following commands are accepted:\r\n"
      "LIST      - display the status of all applications\r\n"
      "STOP #    - stop the application specified\r\n"
      "START #   - start the application specified\r\n"
      "RESTART # - restart the application specified\r\n"
      "KILL #    - stop the application specified, don't wait for users to quit\r\n"
      "KISTART # - restart the application specified, don't wait for users to quit\r\n"
      "QUIT      - close your connection and leave the server\r\n"
      "SHUTDOWN  - shut the entire server down\r\n"
      "HELP      - display this list\r\n"
      "\r\n";

   socket.Write(help.c_str(), GetStringLength<DWORD>(help));
}

void CConfigurationSocketServer::ListApplications(
   IStreamSocket &socket) const
{
   string response = "The following applications are available:\r\n";

   size_t i = 0;

   for (CApplicationCollection::Iterator it = m_applications.Begin(), end = m_applications.End();
      it != end;
      ++it)
   {
      response += ToStringA(i++) + " - " + CStringConverter::TtoA(it->first);

      if (m_applications.ApplicationIsRunning(it->first))
      {
         response += " - running\r\n";
      }
      else
      {
         response += " - stopped\r\n";
      }
   }

   response += "\r\n";

   socket.Write(response.c_str(), GetStringLength<DWORD>(response));
}

bool CConfigurationSocketServer::StopApplication(
   const string &command,
   IStreamSocket &socket,
   const bool kill)
{
   string response;

   bool ok = false;

   if (IsAllDigitsA(command))
   {
      const int appIndex = atoi(command.c_str());

      int i = 0;

      for (CApplicationCollection::Iterator it = m_applications.Begin(), end = m_applications.End();
         !ok && it != end;
         ++it, ++i)
      {
         if (i == appIndex)
         {
            const string application = CStringConverter::TtoA(it->first);

            if (m_applications.StopApplication(it->second->application, kill))
            {
               response = "Application: " + ToStringA(appIndex) + " \"" + application + "\" stopped\r\n";
            }
            else
            {
               response = "Application: " + ToStringA(appIndex) + " \"" + application + "\" was not running\r\n";
            }

            ok = true;
         }
      }

      if (!ok)
      {
         response =
            "Application: " + ToStringA(appIndex) + " not found.\r\n"
            "Type LIST to list available applications\r\n"
            "\r\n";
      }
   }
   else
   {
      response =
         "Error: command format is STOP [appIndex]\r\n"
         "where [appIndex] is the number of the applicatuon\r\n";
         "obtained by a LIST command.\r\n"
         "\r\n";
   }

   socket.Write(response.c_str(), GetStringLength<DWORD>(response));

   return ok;
}

bool CConfigurationSocketServer::StartApplication(
   const string &command,
   IStreamSocket &socket)
{
   string response;

   bool ok = false;

   if (IsAllDigitsA(command))
   {
      const int appIndex = atoi(command.c_str());

      int i = 0;

      for (CApplicationCollection::Iterator it = m_applications.Begin(), end = m_applications.End();
         !ok && it != end;
         ++it, ++i)
      {
         if (i == appIndex)
         {
            const string application = CStringConverter::TtoA(it->first);

            if (m_applications.StartApplication(it->second->application))
            {
               response = "Application: " + ToStringA(appIndex) + " \"" + application + "\" started\r\n";
            }
            else
            {
               response = "Application: " + ToStringA(appIndex) + " \"" + application + "\" was already running\r\n";
            }

            ok = true;
         }
      }

      if (!ok)
      {
         response =
            "Application: " + ToStringA(appIndex) + " not found.\r\n"
            "Type LIST to list available applications\r\n"
            "\r\n";
      }
   }
   else
   {
      response =
         "Error: command format is START [appIndex]\r\n"
         "where [appIndex] is the number of the applicatuon\r\n";
         "obtained by a LIST command.\r\n"
         "\r\n";
   }

   socket.Write(response.c_str(), GetStringLength<DWORD>(response));

   return ok;
}

bool CConfigurationSocketServer::RestartApplication(
   const string &command,
   IStreamSocket &socket,
   const bool kill)
{
   string response;

   bool ok = false;

   if (IsAllDigitsA(command))
   {
      const int appIndex = atoi(command.c_str());

      int i = 0;

      for (CApplicationCollection::Iterator it = m_applications.Begin(), end = m_applications.End();
         !ok && it != end;
         ++it, ++i)
      {
         if (i == appIndex)
         {
            const string application = CStringConverter::TtoA(it->first);

            if (m_applications.RestartApplication(it->second->application, kill))
            {
               response = "Application: " + ToStringA(appIndex) + " \"" + application + "\" restarted\r\n";
            }
            else
            {
               response = "Application: " + ToStringA(appIndex) + " \"" + application + "\" started\r\n";
            }

            ok = true;
         }
      }

      if (!ok)
      {
         response =
            "Application: " + ToStringA(appIndex) + " not found.\r\n"
            "Type LIST to list available applications\r\n"
            "\r\n";
      }
   }
   else
   {
      response =
         "Error: command format is RESTART [appIndex]\r\n"
         "where [appIndex] is the number of the applicatuon\r\n";
         "obtained by a LIST command.\r\n"
         "\r\n";
   }

   socket.Write(response.c_str(), GetStringLength<DWORD>(response));

   return ok;
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static string ProcessCommandString(
   const string &input)
{
   const string::size_type length = input.length() - 2;   // strip the \r\n

   string output;

   output.resize(length);

   string::size_type o = 0;

   for (string::size_type i = 0; i < length; ++i)
   {
      const char c = input[i];

      if (o != 0 && c == 0x08)
      {
         o--;
      }
      else if (o != 0 && isspace(c) && output[o] == ' ')
      {
         // do nothing, collapse multiple space into single
      }
      else
      {
         output[o++] = input[i];
      }
   }

   output.resize(o);

   output = StripLeadingA(output, ' ');

   output = StripTrailingA(output, ' ');

   return output;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ConfigurationSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
