///////////////////////////////////////////////////////////////////////////////
// File: ManagedHost.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2010 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "ManagedHost.h"
#include "UnmanagedHost.h"
#include "IListenForAppDomainUnload.h"

#include "ManagedHostInterface/IManagedHost.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ToString.h"

#include "JetByteTools/Win32Tools/EnvironmentVariables.h"

#include "JetByteTools/COMTools/Utils.h"

#include "JetByteTools/CLRHostingTools/Exception.h"
#include "JetByteTools/CLRHostingTools/CLRPolicyManager.h"
#include "JetByteTools/CLRHostingTools/HostPolicyManager.h"
#include "JetByteTools/CLRHostingTools/CLRHostFactory.h"

#pragma hdrstop

#include <comdef.h>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::ILockableObject;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CStringConverter;

using JetByteTools::Win32::CWin32Exception;
using JetByteTools::Win32::CEnvironmentVariables;

using JetByteTools::CLRHosting::CCLRHost;
using JetByteTools::CLRHosting::CHostPolicyManager;
using JetByteTools::CLRHosting::CCLRHostFactory;
using JetByteTools::CLRHosting::CCLRPolicyManager;

using JetByteTools::COM::SafeAddRef;
using JetByteTools::COM::SafeRelease;

///////////////////////////////////////////////////////////////////////////////
// CManagedHost
///////////////////////////////////////////////////////////////////////////////

CManagedHost::CManagedHost(
   const _tstring &assembly,
   const _tstring &type,
   IListenForAppDomainUnload &domainUnloadListener,
   const bool enableManagedDebuggingAssistants)
   :  m_spClrHost(nullptr),
      m_assembly(assembly),
      m_type(type),
      m_domainUnloadListener(domainUnloadListener),
      m_spCLRControl(0),
      m_spUnmanagedHost(0),
      m_spPolicyManager(0)
{
   Start(enableManagedDebuggingAssistants);

   SetupCLRErrorEscalationPolicy();

   ConnectDefaultAppDomainManager();
}

CManagedHost::~CManagedHost()
{
   try
   {
      if (m_spClrHost.get())
      {
         m_spClrHost->DisableCallbacks(true);
      }

      if (m_spPolicyManager.Get())
      {
         m_spPolicyManager->DisableCallbacks(true);
      }

      for (DomainManagerMap::iterator it = m_domainManagers.begin(); it != m_domainManagers.end(); it++)
      {
         SafeRelease(it->second);
      }

      m_spCLRControl.Release();

      m_spUnmanagedHost.Release();

      m_spPolicyManager.Release();

      if (m_spClrHost.get())
      {
         m_spClrHost->Stop();
      }
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

long CManagedHost::CreateApplicationInNewAppDomain(
   const _tstring &application,
   const _tstring &applicationDirectory,
   const _tstring &sharedDirectory,
   const _tstring &configFile,
   const bool enableShadowCopy)
{
   CSmartManagedHost spHost(GetDefaultManagedHost());

   long appDomainID = 0;

   HRESULT hr = spHost->CreateApplicationInNewAppDomain(
      _bstr_t(application.c_str()),                         // doubles as app domain name for now
      _bstr_t(application.c_str()),
      _bstr_t(applicationDirectory.c_str()),
      _bstr_t(sharedDirectory.c_str()),
      _bstr_t(configFile.c_str()),
      enableShadowCopy,
      &appDomainID);

   if (FAILED(hr))
   {
      throw JetByteTools::CLRHosting::CException(_T("CManagedHost::CreateApplicationInNewAppDomain() - Failed to create new application domain for : \"") + application + _T("\""), hr);
   }

   return appDomainID;
}

IManagedApplication *CManagedHost::StartApplication(
   const long appDomainID,
   const _tstring &assembly,
   const _tstring &type,
   const _tstring &application)
{
   CSmartManagedHost spHost(GetManagedHost(appDomainID));

   IManagedApplication *pApp = 0;

   const HRESULT hr = spHost->Start(
      _bstr_t(assembly.c_str()),
      _bstr_t(type.c_str()),
      _bstr_t(application.c_str()),
      &pApp);

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CManagedHost::StartApplication() - Failed to start application in AppDomain: ") + ToString(appDomainID), hr);

   return pApp;
}

void CManagedHost::RequestApplicationStop(
   const long appDomainID) const
{
   CSmartManagedHost spHost(GetManagedHost(appDomainID));

   const HRESULT hr = spHost->RequestStop();

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CManagedHost::RequestApplicationStop() - Failed to request application to stop in AppDomain: ") + ToString(appDomainID), hr);
}

void CManagedHost::StopApplication(
   const long appDomainID) const
{
   CSmartManagedHost spHost(GetManagedHost(appDomainID));

   const HRESULT hr = spHost->Stop();

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CManagedHost::StopApplication() - Failed to stop application in AppDomain: ") + ToString(appDomainID), hr);
}

void CManagedHost::DestroyAppDomain(
   const long appDomainID)
{
   ILockableObject::Owner lock(m_lock);

   DomainManagerMap::iterator it = m_domainManagers.find(appDomainID);

   if (m_domainManagers.end() != it)
   {
      it->second->Release();

      m_domainManagers.erase(it);

      m_spClrHost->UnloadAppDomain(appDomainID);
   }
}

void CManagedHost::Start(
   const bool enableManagedDebuggingAssistants)
{
   if (m_spClrHost.get())
   {
      throw CException(_T("CManagedHost::Start()"), _T("Already started"));
   }

   if (enableManagedDebuggingAssistants)
   {
      // Use this to enable MDAs, then add an EchoServerCLR.exe.mda.config file ...

      CEnvironmentVariables env;

      env.SetVariable(_T("COMPLUS_MDA"), _T("1"));
   }

   _tstring version = _T("v2.0");

   CSmartCLRRuntimeHost spHost(CCLRHostFactory::GetHostByVersion(version));

   if (!spHost.Get())
   {
      throw CException(_T("CManagedHost::Start()"), _T("Failed to load version \"") + version + _T("\" of the CLR"));
   }

   m_spClrHost.reset(new CCLRHost(spHost.Get(), this, this));

   m_spClrHost->SetAppDomainManagerType(m_assembly, m_type);

   if (!m_spClrHost->StartIfPossible())
   {
      throw CException(_T("CManagedHost::Start()"), _T("Failed to start version \"") + version + _T("\" of the CLR. Already running?"));
   }

   m_spClrHost->RegisterForEventCallback(Event_ClrDisabled);
   m_spClrHost->RegisterForEventCallback(Event_DomainUnload);

   if (enableManagedDebuggingAssistants)
   {
      m_spClrHost->RegisterForEventCallback(Event_MDAFired);
   }

   m_spClrHost->RegisterForEventCallback(Event_StackOverflow);
}

void CManagedHost::SetupCLRErrorEscalationPolicy()
{
   CCLRPolicyManager &policyManager = m_spClrHost->GetPolicyManager();

   policyManager.SetUnhandledExceptionPolicy(eHostDeterminedPolicy);

   policyManager.SetTimeoutAndAction(OPR_AppDomainUnload, 30000, eRudeUnloadAppDomain);
   policyManager.SetTimeoutAndAction(OPR_FinalizerRun, 30000, eUnloadAppDomain);
   policyManager.SetTimeoutAndAction(OPR_ThreadAbort, 30000, eRudeAbortThread);

   policyManager.SetActionOnFailure(FAIL_NonCriticalResource, eAbortThread);
   policyManager.SetActionOnFailure(FAIL_CriticalResource, eUnloadAppDomain);
   policyManager.SetActionOnFailure(FAIL_OrphanedLock, eUnloadAppDomain);
   policyManager.SetActionOnFailure(FAIL_FatalRuntime, eDisableRuntime);
   policyManager.SetActionOnFailure(FAIL_StackOverflow, eRudeUnloadAppDomain);
}

void CManagedHost::ConnectDefaultAppDomainManager()
{
   CSmartManagedHost spHost(GetDefaultManagedHost());

   if (!m_spUnmanagedHost.Get())
   {
      m_spUnmanagedHost = new CUnmanagedHost();
   }

   const HRESULT hr = spHost->InitialiseDefaultAppDomain(m_spUnmanagedHost.Get());

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CManagedHost::ConnectDefaultAppDomainManager()"), hr);
}

void CManagedHost::Stop()
{
   if (m_spClrHost.get())
   {
      m_spClrHost->Stop();
   }

   m_spClrHost.reset(0);
}

CSmartManagedHost CManagedHost::GetManagedHost(
   const long appDomainID) const
{
   {
      ILockableObject::Owner lock(m_lock);

      DomainManagerMap::const_iterator it = m_domainManagers.find(appDomainID);

      if (m_domainManagers.end() != it)
      {
         return CSmartManagedHost(it->second, CSmartManagedHost::CreateNewReference);
      }
   }

   throw CException(_T("CManagedHost::GetManagedHost()"), _T("AppDomain Manager not available for id: ") + ToString(appDomainID));
}

CSmartManagedHost CManagedHost::GetDefaultManagedHost() const
{
   return GetManagedHost(1);
}

HRESULT CManagedHost::GetHostManager(
   REFIID riid,
   void **ppObject)
{
   if (!ppObject)
   {
      return E_POINTER;
   }

   *ppObject = 0;

   HRESULT hr = E_NOINTERFACE;

   if (riid == IID_IHostPolicyManager)
   {
      if (m_spPolicyManager.Get())
      {
         m_spPolicyManager = new CHostPolicyManager(*this);
      }

      m_spPolicyManager.AddRef();

      *ppObject = static_cast<IHostPolicyManager *>(m_spPolicyManager.Get());

      hr = S_OK;
   }

   return hr;
}

HRESULT CManagedHost::SetAppDomainManager(
  DWORD dwAppDomainID,
  IUnknown *pUnkAppDomainManager)
{
   ILockableObject::Owner lock(m_lock);

   DomainManagerMap::iterator it = m_domainManagers.find(dwAppDomainID);

   if (m_domainManagers.end() != it)
   {
      it->second = SafeRelease(it->second);
   }

   IManagedHost *pHost = 0;

   const HRESULT hr = pUnkAppDomainManager->QueryInterface(IID_IManagedHost, (void**)&pHost);

   if (SUCCEEDED(hr))
   {
      m_domainManagers[dwAppDomainID] = pHost;
   }

   return hr;
}

HRESULT CManagedHost::Setup(
   ICLRRuntimeHost *pHost)
{
   ICLRControl *pCLRControl = 0;

   const HRESULT hr = pHost->GetCLRControl(&pCLRControl);

   m_spCLRControl = pCLRControl;

   JetByteTools::CLRHosting::CException::ThrowOnFailure(_T("CManagedHost::Setup() - GetCLRControl"), hr);

   return S_OK;
}

void CManagedHost::OnDefaultAction(
   EClrOperation operation,
   EPolicyAction action)
{
   OutputEx(_T("CManagedHost::OnDefaultAction() - ") + CHostPolicyManager::OperationAsString(operation) + _T(" - ") + CHostPolicyManager::PolicyActionAsString(action));
}

void CManagedHost::OnTimeout(
   EClrOperation operation,
   EPolicyAction action)
{
   OutputEx(_T("CManagedHost::OnTimeout() - ") + CHostPolicyManager::OperationAsString(operation) + _T(" - ") + CHostPolicyManager::PolicyActionAsString(action));
}

void CManagedHost::OnFailure(
   EClrFailure failure,
   EPolicyAction action)
{
   OutputEx(_T("CManagedHost::OnFailure() - ") + CHostPolicyManager::FailureAsString(failure) + _T(" - ") + CHostPolicyManager::PolicyActionAsString(action));
}

void CManagedHost::OnCLRDisabled()
{
   OutputEx(_T("CManagedHost::OnCLRDisabled()"));
}

void CManagedHost::OnDomainUnload(
   const int domainId)
{
   OutputEx(_T("CManagedHost::OnDomainUnload() - ") + ToString(domainId));

   IManagedHost *pDomainManager = 0;

   {
      ILockableObject::Owner lock(m_lock);

      DomainManagerMap::iterator it = m_domainManagers.find(domainId);

      if (it != m_domainManagers.end())
      {
         pDomainManager = it->second;

         m_domainManagers.erase(it);
      }
   }

   if (pDomainManager)
   {
      pDomainManager->Release();
   }

   m_domainUnloadListener.OnAppDomainUnload(domainId);
}

void CManagedHost::OnMDAEvent(
   const _tstring &title,
   const _tstring &message)
{
   OutputEx(_T("CManagedHost::OnMDAEvent() - ") + title + _T(" - ") + message);
}

void CManagedHost::OnStackOverflow()
{
   // Note that if you are running in the debugger when you get a Stack Overflow
   // event then the process seems to hang after this debug line is displayed.
   // No idea why... Probably a COM threading deadlock issue either due to
   // something we're doing or the interaction between the managed->unmanaged
   // transition and the managed debug api.
   // Note that this is only a problem if the debugger type is set to 'mixed',
   // it seems fine if the debugger is set to only debug native code.

   OutputEx(_T("CManagedHost::OnStackOverflow()"));
}

void CManagedHost::OnException()
{
   Output(_T("OnException()"));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ManagedHost.cpp
///////////////////////////////////////////////////////////////////////////////
