using System;
using System.Threading;
using System.Reflection;
using System.Collections.Generic;

using EchoServerCLRManagedHostInterfaces;

namespace EchoServerCLR
{
   class ManagedHost : AppDomainManager, IManagedHost
   {
      private bool defaultAppDomain = true;

      private string appDomainName = "Default";

      private IManagedApplicationControl application;

      private IUnmanagedHost host;

      public override void InitializeNewDomain(AppDomainSetup appDomainInfo)
      {
         // let the unmanaged host know about us
         InitializationFlags = AppDomainManagerInitializationOptions.RegisterWithHost;
      }

      void IManagedHost.InitialiseDefaultAppDomain(IUnmanagedHost host)
      {
         if (!defaultAppDomain)
         {
            throw new InvalidOperationException("Should only be called on the default application domain, not on \"" + appDomainName + "\"");
         }

         if (this.host != null)
         {
            throw new InvalidOperationException("Already initialised");
         }

         this.host = host;

         SetAppDomainEventHandlers(AppDomain.CurrentDomain, false);
      }

      int IManagedHost.CreateApplicationInNewAppDomain(
         string appDomainName, 
         string applicationName, 
         string applicationDirectory, 
         string sharedDirectory,
         string configFile,
         int enableShadowCopy)
      {
         if (!defaultAppDomain)
         {
            throw new InvalidOperationException("Should only be called on the default application domain, not on \"" + this.appDomainName + "\"");
         }

         if (host == null)
         {
            throw new InvalidOperationException("Not initialised, call InitialiseDefaultAppDomain() first!");
         }

         AppDomainSetup setup = new AppDomainSetup();

         setup.ApplicationName = applicationName;
         setup.ApplicationBase = ".";
         setup.PrivateBinPathProbe = "true";
         setup.PrivateBinPath = applicationDirectory + ";" + sharedDirectory;
         setup.ConfigurationFile = configFile;

         if (enableShadowCopy == 1)
         {
            setup.ShadowCopyFiles = "true";
            setup.CachePath = ".\\Cache\\" + applicationDirectory;
         }

         // v4.0 features
         //setup.AppDomainManagerAssembly = "";
         //setup.AppDomainManagerType = "";

         AppDomain appDomain = AppDomain.CreateDomain(appDomainName, null, setup);

         ManagedHost manager = (ManagedHost)appDomain.DomainManager;

         manager.SetupAppDomain(appDomainName, appDomain, host);

         return appDomain.Id;
      }

      IManagedApplication IManagedHost.Start(
         string assemblyName,
         string typeName,
         string applicationName)
      {
         if (defaultAppDomain)
         {
            throw new InvalidOperationException("Should only be called on the application domain that was created for application \"" + application + "\", not on the default application domain");
         }

         if (appDomainName != applicationName)
         {
            throw new InvalidOperationException("Should only be called on the application domain that was created for application \"" + application + "\", not on \"" + appDomainName + "\"");
         }

         Assembly assembly = Assembly.Load(assemblyName);

         object obj = assembly.CreateInstance(typeName);

         if (obj == null)
         {
            throw new TypeLoadException("Object \"" + typeName + "\" not found in assembly \"" + assemblyName + "\"");
         }

         this.application = (IManagedApplicationControl)obj;

         return application.OnApplicationStart(applicationName, host);
      }

      void IManagedHost.RequestStop()
      {
         if (defaultAppDomain)
         {
            throw new InvalidOperationException("Should not be called on the default application domain");
         }

         if (application == null)
         {
            throw new InvalidOperationException("Should only be called if an application has been started");
         }

         application.OnApplicationStopRequested();
      }

      void IManagedHost.Stop()
      {
         if (defaultAppDomain)
         {
            throw new InvalidOperationException("Should not be called on the default application domain");
         }

         if (application == null)
         {
            throw new InvalidOperationException("Should only be called if an application has been started");
         }

         application.OnApplicationStop();
      }

      private void SetupAppDomain(
         string appDomainName,
         AppDomain appDomain,
         IUnmanagedHost host)
      {
         defaultAppDomain = false;

         this.appDomainName = appDomainName;
         this.host = host;

         SetAppDomainEventHandlers(appDomain, true);
      }

      private void SetAppDomainEventHandlers(
         AppDomain appDomain,
         bool includeExceptions)
      {
         // We MUST route the events to the app domain manager of the app domain that is
         // generating the events as there are marshalling issues otherwise.

         ManagedHost manager = (ManagedHost)appDomain.DomainManager;

         if (includeExceptions)
         {
            appDomain.UnhandledException += new UnhandledExceptionEventHandler(manager.UnhandledExceptionHandler);
         }

         appDomain.AssemblyLoad += new AssemblyLoadEventHandler(manager.AssemblyLoadHandler);
         appDomain.AssemblyResolve += new ResolveEventHandler(manager.AssemblyResolveHandler);
         appDomain.DomainUnload += new EventHandler(manager.DomainUnloadHandler);
         //appDomain.FirstChanceException - 4.0 only
         appDomain.ReflectionOnlyAssemblyResolve += new ResolveEventHandler(manager.ReflectionOnlyAssemblyResolveHandler);
         appDomain.ResourceResolve += new ResolveEventHandler(manager.ResourceResolveHandler);
         appDomain.TypeResolve += new ResolveEventHandler(manager.TypeResolveHandler);
      }

      private void LogMessage(
         string message)
      {
         host.LogMessage(AppDomain.CurrentDomain.Id + ": " + appDomainName + ": " + message);
      }

      public void AssemblyLoadHandler(
         Object sender,
         AssemblyLoadEventArgs args)
      {
         LogMessage("AssemblyLoadEvent: " + args.LoadedAssembly.FullName);
      }

      public Assembly AssemblyResolveHandler(
         Object sender,
         ResolveEventArgs args)
      {
         LogMessage("AssemblyResolveEvent: " + args.Name);

         return null;
      }

      public void DomainUnloadHandler(
         Object sender,
         EventArgs e)
      {
         LogMessage("domain unloaded");
      }

      public Assembly ReflectionOnlyAssemblyResolveHandler(
         Object sender,
         ResolveEventArgs args)
      {
         LogMessage("ReflectionOnlyAssemblyResolveHandler: " + args.Name);

         return null;
      }

      public Assembly ResourceResolveHandler(
         Object sender,
         ResolveEventArgs args)
      {
         LogMessage("ResourceResolveHandler: " + args.Name);

         return null;
      }

      public Assembly TypeResolveHandler(
         Object sender,
         ResolveEventArgs args)
      {
         LogMessage("TypeResolveHandler: " + args.Name);

         return null;
      }

      public void UnhandledExceptionHandler(
         object sender, 
         UnhandledExceptionEventArgs e)
      {
         LogMessage("Unhandled exception: " + e.ExceptionObject.ToString());
      }
   }
}
