///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2010 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "ConfigurationSocketServer.h"
#include "SocketServer.h"
#include "ApplicationCollection.h"
#include "IOPool.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/CommandLine.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/ServerCollection.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"
#include "JetByteTools/SocketTools/FullAddress.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;
using JetByteTools::Core::GetCurrentDirectory;

using JetByteTools::Socket::CServerCollection;
using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;
using JetByteTools::Socket::IIPAddress;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   ::CoInitializeEx(0, COINIT_MULTITHREADED);

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("EchoServerCLR4.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CCommandLine commandLine(argc, argv, _T("EchoServerCLR4"), CCommandLine::TCPServer);

         if (commandLine.Parse())
         {
            CIOPool pool(commandLine.NumberOfIOThreads());

            pool.Start();

            CSequencedStreamSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            const ListenBacklog listenBacklog = commandLine.ListenBacklog();

            CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

            const _tstring assembly = _T("ManagedHost, Version=1.0.0.0, PublicKeyToken=24e328fd91cb8ee6");
            const _tstring type = _T("EchoServerCLR.ManagedHost");

            const bool enableManagedDebuggingAssistants = true;

            const _tstring baseDirectory = GetCurrentDirectory();

            CApplicationCollection applications(
               baseDirectory,
               assembly,
               type,
               enableManagedDebuggingAssistants);

            CServerCollection servers;

            IIPAddress::Port port = commandLine.Port();

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            {

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CConfigurationSocketServer(
                  applications,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               const _tstring applicationName = _T("Application 1");

               const _tstring applicationDir = _T("Apps\\Application1");

               const _tstring assemblyFileName = _T("ManagedApplication");

               applications.AddApplication(
                  applicationName,
                  applicationDir,
                  _T("Shared; Apps\\Shared"),
                  applicationDir + _T("\\") + assemblyFileName + _T(".dll.config"),
                  CApplicationCollection::DisableShadowCopy,
                  assemblyFileName + _T(", Version=1.0.0.0, PublicKeyToken=24e328fd91cb8ee6"),
                  _T("EchoServerCLR.ManagedApplication"));

               applications.StartApplication(
                  applicationName);

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CSocketServer(
                  applications,
                  applicationName,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               const _tstring applicationName = _T("Application 2");

               const _tstring applicationDir = _T("Apps\\Application2");

               const _tstring assemblyFileName = _T("ManagedApplication");

               applications.AddApplication(
                  applicationName,
                  applicationDir,
                  _T("Shared; Apps\\Shared"),
                  applicationDir + _T("\\") + assemblyFileName + _T(".dll.config"),
                  CApplicationCollection::EnableShadowCopy,
                  assemblyFileName + _T(", Version=1.0.0.0, PublicKeyToken=24e328fd91cb8ee6"),
                  _T("EchoServerCLR.ManagedApplication"));

               applications.StartApplication(
                  applicationName);

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CSocketServer(
                  applications,
                  applicationName,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               const _tstring applicationName = _T("Application 3");

               const _tstring applicationDir = _T("Apps\\Application3");

               const _tstring assemblyFileName = _T("ManagedApplication");

               applications.AddApplication(
                  applicationName,
                  applicationDir,
                  _T("Shared; Apps\\Shared"),
                  applicationDir + _T("\\") + assemblyFileName + _T(".dll.config"),
                  CApplicationCollection::EnableAutoRestart,
                  assemblyFileName + _T(", Version=1.0.0.0, PublicKeyToken=24e328fd91cb8ee6"),
                  _T("EchoServerCLR.ManagedApplication"));

               applications.StartApplication(
                  applicationName);

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CSocketServer(
                  applications,
                  applicationName,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               const _tstring applicationName = _T("Application 4");

               const _tstring applicationDir = _T("Apps\\Application4");

               const _tstring assemblyFileName = _T("ManagedApplication");

               applications.AddApplication(
                  applicationName,
                  applicationDir,
                  _T("Shared; Apps\\Shared"),
                  applicationDir + _T("\\") + assemblyFileName + _T(".dll.config"),
                  CApplicationCollection::EnableAutoForceRestart,
                  assemblyFileName + _T(", Version=1.0.0.0, PublicKeyToken=24e328fd91cb8ee6"),
                  _T("EchoServerCLR.ManagedApplication"));

               applications.StartApplication(
                  applicationName);

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CSocketServer(
                  applications,
                  applicationName,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            try
            {
               servers.Start();

               servers.StartAcceptingConnections();

            shutdownHandler.SetServerControl(servers, commandLine.EventNamePrefix());

               shutdownHandler.WaitForShutdownRequest();
            }
            catch(const CException &e)
            {
               OutputEx(_T("Exception: ") + e.GetDetails());

               result = 1;
            }
            catch(...)
            {
               OutputEx(_T("Unexpected exception"));

               result = 1;
            }

            servers.BeginShutdown();

            applications.StopAllApplications();

            servers.WaitForShutdownToComplete();

            pool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
